#define STATIC static

#define OK		1

typedef unsigned long U32;
typedef unsigned char U8; 

extern void HaltFunc();


#define NFCONF		((volatile unsigned *)0x4E000000)		//NAND Flash configuration
#define NFCONT		((volatile unsigned *)0x4E000004)      //NAND Flash control
#define NFCMD		((volatile unsigned *)0x4E000008)      //NAND Flash command
#define NFADDR		((volatile unsigned *)0x4E00000C)      //NAND Flash address
#define NFDATA		((volatile unsigned *)0x4E000010)      //NAND Flash data
#define NFDATA8	((volatile unsigned char *)0x4E000010)     //NAND Flash data
// #define NFDATA		(0x4E000010)      //NAND Flash data address
#define NFMECCD0	((volatile unsigned *)0x4E000014)      //NAND Flash ECC for Main Area
#define NFMECCD1	((volatile unsigned *)0x4E000018)
#define NFSECCD	((volatile unsigned *)0x4E00001C)		//NAND Flash ECC for Spare Area
#define NFSTAT		((volatile unsigned *)0x4E000020)		//NAND Flash operation status
#define NFESTAT0	((volatile unsigned *)0x4E000024)
#define NFESTAT1	((volatile unsigned *)0x4E000028)
#define NFMECC0	((volatile unsigned *)0x4E00002C)
#define NFMECC1	((volatile unsigned *)0x4E000030)
#define NFSECC		((volatile unsigned *)0x4E000034)
#define NFSBLK		((volatile unsigned *)0x4E000038)		//NAND Flash Start block address
#define NFEBLK		((volatile unsigned *)0x4E00003C)		//NAND Flash End block address


#define NF_MECC_UnLock()    {*NFCONT&=~(1<<5);}
#define NF_MECC_Lock()      {*NFCONT|=(1<<5);}

#define NF_CMD(cmd)			{*NFCMD=cmd;}
#define NF_ADDR(addr)		{*NFADDR=addr;}
#define NF_nFCE_L()			{*NFCONT&=~(1<<1);}
#define NF_nFCE_H()			{*NFCONT|=(1<<1);}
#define NF_RSTECC()			{*NFCONT|=(1<<4);}
#define NF_RDDATA() 		(*NFDATA)
#define NF_RDDATA8() 		((*(volatile unsigned char*)0x4E000010) )

#define NF_WRDATA(data) 	{*NFDATA=data;}

#define NF_WAITRB()    		{while(!((*NFSTAT)&(1<<1)));}
	   						 //wait tWB and check F_RNB pin.
// RnB Signal
#define NF_CLEAR_RB()    	{*NFSTAT |= (1<<2);}	// Have write '1' to clear this bit.
#define NF_DETECT_RB()    	{while(!((*NFSTAT)&(1<<2)));}

#define TACLS		0	// 1-clk(0ns)
#define TWRPH0		6	// 3-clk(25ns)
#define TWRPH1		0	// 1-clk(10ns)  //TACLS+TWRPH0+TWRPH1>=50ns

STATIC void _Nand_Minimal_Reset(void);
STATIC int _Nand_Minimal_IsBadBlockPage512(U32 block);
STATIC int _Nand_Minimal_IsBadBlockPage2048(U32 block);
STATIC U32 _Nand_Minimal_NF8_CheckId(void);
STATIC int _Nand_Minimal_Init(void);
STATIC int _Nand_Minimal_IsBadBlock(U32 block,int bLargeBlock);

STATIC void _Nand_Minimal_Reset(void)
{
    volatile int i;

	NF_nFCE_L();
	NF_CLEAR_RB();
	for (i=0; i<10; i++);
	NF_CMD(0xFF);	//reset command
	NF_DETECT_RB();
	NF_nFCE_H();

}


STATIC int _Nand_Minimal_IsBadBlock(U32 block,int bLargeBlock)
{
	switch(bLargeBlock) {
	case 0:
		return _Nand_Minimal_IsBadBlockPage512(block);
	case 1:
		return _Nand_Minimal_IsBadBlockPage2048(block);
	}
	return 0;

}


STATIC U32 _Nand_Minimal_NF8_CheckId(void)
{
	int i;
	U8 Mid, Did, DontCare, id4th;

	NF_nFCE_L();

 	NF_CMD(0x90);
	NF_ADDR(0x0);
	for (i=0; i<100; i++);

	Mid 		= NF_RDDATA8();
	Did   		= NF_RDDATA8();
	DontCare 	= NF_RDDATA8();
	id4th 		= NF_RDDATA8();

	NF_nFCE_H();

	switch(Did) {
	case 0x76:
		return 0;
	case 0xF1:
	case 0xD3:
	case 0xDA:
	case 0xDC:
		return 1;
	default:
		HaltFunc();
		//Uart_SendString("Unknown NAND type\r\n");
		for(;;);
	}


}

STATIC int _Nand_Minimal_Init(void)
{
	*NFCONF = (TACLS<<12)|(TWRPH0<<8)|(TWRPH1<<4)|(0<<0);
	*NFCONT = (0<<13)|(0<<12)|(0<<10)|(0<<9)|(0<<8)|(0<<6)|(0<<5)|(1<<4)|(1<<1)|(1<<0);
	*NFSTAT = 0;

	_Nand_Minimal_Reset();

	return _Nand_Minimal_NF8_CheckId();
}


STATIC void __RdPage512(U8 *buf)
{
	unsigned i;
	for (i = 0; i < 512; i++) {
		buf[i] =  NF_RDDATA8();
	}
}







STATIC int _Nand_Minimal_IsBadBlockPage512(U32 block)
{
    U32 	Page;
	U8 		BAD;

    Page = (block<<8);

    _Nand_Minimal_Reset();

    NF_nFCE_L();
	NF_CLEAR_RB();
	NF_CMD(0x50);
  	NF_ADDR(5);
    NF_ADDR(Page&0xff);
    NF_ADDR((Page>>8)&0xff);
    NF_ADDR((Page>>16)&0xff);
    NF_DETECT_RB();

    BAD = NF_RDDATA8();

    NF_nFCE_H();

    return BAD != 0xff;

}

STATIC int _Nand_Minimal_IsBadBlockPage2048(U32 block)
{
    U32 	Page;
	U8 		BAD;

   	Page = block << 6;
   	NF_nFCE_L();
    NF_CLEAR_RB();
   	NF_CMD(0x00);
	NF_ADDR(0);
	NF_ADDR((2048>>8)&0xff);
    NF_ADDR(Page&0xff);
    NF_ADDR((Page>>8)&0xff);
    NF_ADDR((Page>>16)&0xff);
    NF_CMD(0x30);
   	NF_DETECT_RB();

   	BAD = NF_RDDATA8();

   	NF_nFCE_H();

    return BAD != 0xff;
}


STATIC int _Nand_Minimal_ReadSectorPage512(U32 sector, U8 *buffer)
{

    _Nand_Minimal_Reset();

	NF_nFCE_L();

	NF_CLEAR_RB();
	NF_CMD(0x00);

    NF_ADDR(0x00);
	NF_ADDR(sector&0xff);
	NF_ADDR((sector>>8)&0xff);
	NF_ADDR((sector>>16)&0xff);

	NF_DETECT_RB();

    __RdPage512(buffer);

	NF_nFCE_H();


    return 1;
}

STATIC int _Nand_Minimal_ReadSectorPage2048(U32 sector, U8 *buffer)
{
	U32 page, data_addr;

    _Nand_Minimal_Reset();

	NF_nFCE_L();

	NF_CLEAR_RB();
	NF_CMD(0x00);						// Read command

	page = sector/4;
	data_addr = 512 *(sector%4);
	NF_ADDR(data_addr&0xff);
	NF_ADDR((data_addr>>8)&0xff);
	NF_ADDR(page&0xff);
	NF_ADDR((page>>8)&0xff);
	NF_ADDR((page>>16)&0xff);
	NF_CMD(0x30);


	NF_DETECT_RB();

    __RdPage512(buffer);

	NF_nFCE_H();

   	return OK;
}


STATIC int _Nand_Minimal_ReadSector(U32 sector, U8 *buffer, int bLargeBlock)
{
	switch(bLargeBlock) {
	case 0:
		return _Nand_Minimal_ReadSectorPage512(sector, buffer);
	case 1:
		return _Nand_Minimal_ReadSectorPage2048(sector, buffer);
	}
	return 0;
}


void Nand_Minimal_CopyProgramFromNand(void)
{
	const int bLargeBlock = _Nand_Minimal_Init();
	const unsigned StartSector = 0, StopSector = 5120;
	const unsigned byte_sector_shift = 9;
	const unsigned sector_block_shift = bLargeBlock ? 8 : 5;
	unsigned char *RAM = (unsigned char *)0x30000000;
	unsigned Sector, GoodSector;

/*	HaltFunc(); */ 

	for (Sector = StartSector, GoodSector = Sector; Sector < StopSector; Sector ++, GoodSector++) {

		// begin of a block
		if (GoodSector & ( (1 << sector_block_shift) - 1 ) == 0) {
			// found a good block
			for (;;) {
				if (!_Nand_Minimal_IsBadBlock(GoodSector >> sector_block_shift, bLargeBlock)) {
					// Is good Block
					break;
				}
				// try next block
				GoodSector += (1 << sector_block_shift);
			}
		}
		_Nand_Minimal_ReadSector(GoodSector, RAM + (Sector << byte_sector_shift ), bLargeBlock);
	}

}


